// Copyright 2019 cruzawl developers
// Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.

import 'dart:async';
import 'dart:math';
import 'dart:typed_data';

import 'package:collection/collection.dart';

typedef StringCallback = void Function(String);
typedef VoidCallback = void Function();

Future<void> voidResult() async {}

/// Optionally [debugPrint] failed assertions in-app.
bool checkEquals(dynamic x, dynamic y, [StringCallback debugPrint]) {
  bool equals = x == y;
  if (debugPrint != null && !equals) debugPrint('assertion failure: $x != $y');
  assert(x == y, '$x != $y');
  return equals;
}

/// Returns true if [x] and [y] are equivalent.
bool equalUint8List(Uint8List x, Uint8List y) {
  if (x.length != y.length) return false;
  for (int i = 0; i < x.length; ++i) {
    if (x[i] != y[i]) return false;
  }
  return true;
}

/// Returns [n] random bytes.
Uint8List randBytes(int n) {
  final Random generator = Random.secure();
  final Uint8List random = Uint8List(n);
  for (int i = 0; i < random.length; i++) {
    random[i] = generator.nextInt(255);
  }
  return random;
}

/// Splits [debugPrint] into separate calls of [maxPrintLength].
void debugPrintLong(Object object, StringCallback debugPrint) async {
  const int maxPrintLength = 1000;
  if (object == null || object.toString().length <= maxPrintLength) {
    debugPrint(object);
  } else {
    final String text = object.toString();
    int startIndex = 0, endIndex = maxPrintLength;
    int remainingLength = text.length;
    while (endIndex < text.length) {
      debugPrint(text.substring(startIndex, endIndex));
      endIndex += maxPrintLength;
      startIndex += maxPrintLength;
      remainingLength -= maxPrintLength;
    }
    if (remainingLength > 0) debugPrint(text.substring(startIndex));
  }
}

/// Saves exception details for presentation.
class ErrorDetails {
  dynamic exception;
  StackTrace stack;
  ErrorDetails({this.exception, this.stack});
}

/// A proto-[Set] implemented as a [List] sorted by [compare].
class SortedListSet<T> {
  List<T> data;
  int Function(T, T) compare;
  SortedListSet(this.compare, this.data);

  bool get isEmpty => data.isEmpty;
  int get length => data.length;
  T get first => data.first;
  T get last => data.last;

  void clear() => data.clear();

  bool add(T value, {bool overwrite = true}) {
    int index = lowerBound(data, value, compare: compare);
    if (index < data.length && compare(data[index], value) == 0) {
      if (overwrite) data[index] = value;
      return false;
    } else {
      data.insert(index, value);
      return true;
    }
  }

  bool remove(T value) {
    int index = binarySearch(data, value, compare: compare);
    if (index < 0) return false;
    data.removeAt(index);
    return true;
  }

  T find(T value) {
    int index = binarySearch(data, value, compare: compare);
    return index < 0 ? null : data[index];
  }
}
